#include "chain.h"

#include "effectfactory.h"
#include "exceptionfactory.h"

#include <iostream>

Chain::Chain(tinyxml2::XMLElement* chainRoot, const int blockSize)
{
    m_chainRoot = chainRoot;
    m_data.blockSize = blockSize;
}

Chain::~Chain()
{

}

void Chain::init()
{
    setupData();
    setupProcesses();
}

void Chain::setupData()
{
    if (m_chainRoot->Attribute("inFile"))
        m_data.effectData.inFilePath = m_chainRoot->Attribute("inFile");
    else
        throw ExceptionFactory::create(m_chainRoot, ChainError::Missing_SF_INFILE);

    if (m_chainRoot->Attribute("outFile"))
        m_data.effectData.outFilePath = m_chainRoot->Attribute("outFile");
    else
        throw ExceptionFactory::create(m_chainRoot, ChainError::Missing_SF_OUTFILE);
    
    if (m_data.effectData.inFilePath == m_data.effectData.outFilePath)
        throw ExceptionFactory::create(m_chainRoot, ChainError::PathError_SF_INFILE_EQ_OUTFILE);

    m_inFile = SndfilePtr(new SndfileHandle(m_data.effectData.inFilePath, SFM_READ));

    if (m_inFile->error() != SF_ERR_NO_ERROR)
        throw ExceptionFactory::create(m_chainRoot, ChainError::FailedToOpen_SF_INFILE);

    m_data.effectData.numFrames = m_inFile->frames();
    m_data.effectData.numChannels = m_inFile->channels();
    m_data.effectData.format = m_inFile->format();
    m_data.effectData.sampleRate = m_inFile->samplerate();


    m_data.bodyFrames = m_data.effectData.numFrames;
    m_data.bodyBlocks = m_data.bodyFrames / m_data.blockSize;
    m_data.bodyFrameRem = m_data.bodyFrames - m_data.bodyBlocks * m_data.blockSize;

    // Tail data gets worked out after processes are added
    m_data.tailFrames = 0;
    m_data.tailBlocks = 0;
    m_data.tailFrameRem = 0;

    m_audioBuffer = AudioBuffer_uPtr(new AudioBuffer(m_data.effectData.numChannels, m_data.blockSize));
    m_interBuffer = std::unique_ptr<float>(new float[m_data.effectData.numChannels * m_data.blockSize]);
}

void Chain::setupProcesses()
{
    tinyxml2::XMLElement* p = m_chainRoot->FirstChildElement();
    while (p != nullptr)
    {
        if (strcmp("PROCESS", p->Name()) == 0)
        {
            addProcess(p);
        }
        p = p->NextSiblingElement();
    }
}

void Chain::addProcess(tinyxml2::XMLElement* processRoot)
{
    EffectPtr newProcess(EffectFactory::instance().createByName(processRoot->Attribute("name"), m_data.effectData));

    if (newProcess != nullptr)
    {
        tinyxml2::XMLElement* paramNode = processRoot->FirstChildElement();
        while (paramNode != nullptr)
        {
            ParamPtr p = newProcess->getParam(paramNode->Attribute("name"));
            
            if (p == nullptr)
                throw ExceptionFactory::create(paramNode, ChainError::DoesNotExist_PARAM);

            p->clear();

            if (paramNode->Attribute("type") == nullptr || strcmp(paramNode->Attribute("type"), "static") == 0)
            {
                float val;
                paramNode->QueryFloatText(&val); // NOTE: returns tinyxml2::XML_CAN_NOT_CONVERT_TEXT if failed
                newProcess->getParam(paramNode->Attribute("name"))->addPoint(0, val);
            }
            else
            {
                tinyxml2::XMLElement* bpNode = paramNode->FirstChildElement();
                
                if (bpNode == nullptr)
                    throw ExceptionFactory::create(paramNode, ChainError::Missing_BREAKPOINTS);

                while (bpNode != nullptr)
                {
                    if (strcmp("BREAKPOINT", bpNode->Name()) == 0) // make sure we've got a BREAKPOINT element
                    {
                        float time, value;

                        // should bomb if user attempted to set breakpoints but failed
                        if (bpNode->QueryFloatAttribute("time", &time) == tinyxml2::XML_NO_ATTRIBUTE)
                            throw ExceptionFactory::create(paramNode, ChainError::Missing_BREAKPOINT_TIME);

                        if (bpNode->QueryFloatAttribute("value", &value) == tinyxml2::XML_NO_ATTRIBUTE)
                            throw ExceptionFactory::create(paramNode, ChainError::Missing_BREAKPOINT_VALUE);

                        newProcess->getParam(paramNode->Attribute("name"))->addPoint(time, value);
                    }
                    else
                    {
                        std::cerr << "WARNING: Process \"" << processRoot->Attribute("name") << "\", PARAM " << paramNode->Name() << " - found unknown element \"" << bpNode->Name() << std::endl;
                    }
                    bpNode = bpNode->NextSiblingElement();
                }
            }
            paramNode = paramNode->NextSiblingElement();
        }

        if (newProcess->addsTail())
            newProcess->reinit();

        m_processes.push_back(newProcess);
    }
    else
    {
        throw ExceptionFactory::create(processRoot, ChainError::DoesNotExist_PROCESS);
    }
}

void Chain::calcTail()
{
    int max = 0;
    for (auto& process : m_processes)
    {
        if (process->numExtraFrames() > max)
            max = process->numExtraFrames();
    }

    m_data.tailFrames = max;
    m_data.tailBlocks = m_data.tailFrames / m_data.blockSize;
    m_data.tailFrameRem = m_data.tailFrames - m_data.tailBlocks * m_data.blockSize;
}


void Chain::processBody(const unsigned int numBlocks, const unsigned int numFrames)
{
    for (int block = 0; block < numBlocks; ++block)
    {
        m_inFile->readf(m_interBuffer.get(), numFrames);
        AudioBuffer::deinterleave(m_interBuffer.get(), m_audioBuffer->get(), m_data.effectData.numChannels, numFrames);

        for (auto& p : m_processes)
        {
            p->process(m_audioBuffer->get(), m_audioBuffer->get(), m_data.effectData.numChannels, numFrames);
        }

        AudioBuffer::interleave(m_audioBuffer->get(), m_interBuffer.get(), m_data.effectData.numChannels, numFrames);
        m_outFile->writef(m_interBuffer.get(), numFrames);
    }
}

void Chain::processTail(const unsigned int numBlocks, const unsigned int numFrames)
{
    for (int block = 0; block < numBlocks; ++block)
    {
        m_audioBuffer->clear();
        for (auto& p : m_processes)
        {
            p->process(m_audioBuffer->get(), m_audioBuffer->get(), m_data.effectData.numChannels, numFrames);
        }

        AudioBuffer::interleave(m_audioBuffer->get(), m_interBuffer.get(), m_data.effectData.numChannels, numFrames);
        m_outFile->writef(m_interBuffer.get(), numFrames);
    }
}

void Chain::process()
{
    m_outFile = SndfilePtr(new SndfileHandle(m_data.effectData.outFilePath, SFM_WRITE, m_data.effectData.format, m_data.effectData.numChannels, m_data.effectData.sampleRate));

    if (m_outFile->error() != SF_ERR_NO_ERROR)
        throw ExceptionFactory::create(m_chainRoot, ChainError::FailedToOpen_SF_OUTFILE);

    std::cout << "." << std::flush;
    processBody(m_data.bodyBlocks, m_data.blockSize);
    std::cout << "." << std::flush;
    processBody(1, m_data.bodyFrameRem);
    std::cout << "." << std::flush;
    calcTail();
    std::cout << "." << std::flush;
    if (m_data.tailFrames > 0)
    {
        std::cout << "." << std::flush;
        processTail(m_data.tailBlocks, m_data.blockSize);
        std::cout << "." << std::flush;
        processTail(1, m_data.tailFrameRem);
        std::cout << "." << std::flush;
    }
}