#ifndef LINK_EFFECT_H
#define LINK_EFFECT_H

#include <string>
#include <vector>
#include <memory>

#include "parameter.h"

#include <stdexcept>
#include <exception>

class EffectException : public std::runtime_error
{
public:
    EffectException(const char* what) : std::runtime_error(what) {}
};

// struct of necessary info for every effect - all instances (probably just one?) should be owned by Engine and passed as creation arg
// allows things like Split to be defined without extra XML specification
struct EffectData
{
    std::string inFilePath;
    std::string outFilePath;
    int numChannels = 0;
    int numFrames = 0;
    int sampleRate = 44100;
    int format = 0;
};

class Effect
{
    std::vector<ParamPtr> m_paramList;
    bool m_addsTail;

protected:
    const EffectData& m_data;
    int m_numExtraFrames;
    // only subclasses can add parameters
    void addParam(const std::string name, const float minValue, const float maxValue, const float initValue)
    {
        ParamPtr tmp(new Parameter(name, minValue, maxValue, initValue));
        m_paramList.push_back(tmp);
    }

    // only subclasses can set addsTail
    void addsTail(bool value) { m_addsTail = value; }

    const std::vector<ParamPtr>& getParamList() const
    {
        return m_paramList;
    }

public:

    Effect(const EffectData& data) : m_numExtraFrames(0), m_addsTail(false), m_data(data) {}
    virtual ~Effect() {}
    virtual void process(float** in, float** out, const int numChannels, const int numFrames) = 0;

    // Returns nullptr if not found - caller must check
    ParamPtr getParam(const std::string& name)
    {
        ParamPtr tmp = nullptr;
        for (const auto& param : m_paramList)
        {
            // if found, set tmp and break out of loop
            if (param->name() == name)
            {
                tmp = param;
                break;
            }
        }

        return tmp;
    }

    // for recalcuating extraBlocks after changing param values
    virtual void reinit() {};

    // check if adds a tail
    bool addsTail() const { return m_addsTail; }

    const int numExtraFrames() const { return m_numExtraFrames; }

};

typedef std::shared_ptr<Effect> EffectPtr;

#endif // LINK_EFFECT_H