#ifndef LINK_EXCEPTIONFACTORY_H
#define LINK_EXCEPTIONFACTORY_H

#include <tinyxml2.h>
#include <exception>
#include <sstream>

enum class ChainError
{
    FailedToOpen_SF_INFILE,
    FailedToOpen_SF_OUTFILE,
    PathError_SF_INFILE_EQ_OUTFILE,
    DoesNotExist_PROCESS,
    DoesNotExist_PARAM,
    Missing_SF_INFILE,
    Missing_SF_OUTFILE,
    Missing_BREAKPOINTS,
    Missing_BREAKPOINT_TIME,
    Missing_BREAKPOINT_VALUE
};

enum class EngineError
{
    FailedToOpen_XML,
    Element_ROOT_MISSING,
    Element_LINK_MISSING
};

class ChainException : public std::runtime_error
{
public:
    ChainException(const char* what) : std::runtime_error(what) {}
};

class EngineException : public std::runtime_error
{
public:
    EngineException(const char* what) : std::runtime_error(what) {}
};


class ExceptionFactory
{
public:
    static ChainException create(tinyxml2::XMLElement* errorNode, ChainError errorCode)
    {
        std::stringstream ss;
        ss << "ERROR IN CHAIN: ";
        switch (errorCode)
        {
        case(ChainError::FailedToOpen_SF_INFILE) :
            ss << "cannot open inFile - " << errorNode->Attribute("inFile");
            break;
        case(ChainError::FailedToOpen_SF_OUTFILE) :
            ss << "cannot open outFile - " << errorNode->Attribute("outFile");
            break;
        case(ChainError::PathError_SF_INFILE_EQ_OUTFILE) :
            ss << "inFile '" << errorNode->Attribute("inFile") << "' equals outFile '" << errorNode->Attribute("outFile") << "' - inFile would be overwritten, cannot continue..";
            break;
        case(ChainError::DoesNotExist_PROCESS) :
            ss << "PROCESS \"" << errorNode->Attribute("name") << "\" - does not exist";
            break;
        case(ChainError::DoesNotExist_PARAM) :
            ss << "PARAM does not exist\n" << "\nPROCESS: " << errorNode->Parent()->ToElement()->Attribute("name") << "\nPARAM: " << errorNode->Attribute("name");
            break;
        case(ChainError::Missing_SF_INFILE) :
            ss << "no inFile provided";
            break;
        case(ChainError::Missing_SF_OUTFILE) :
            ss << "no outFile provided";
            break;
        case(ChainError::Missing_BREAKPOINTS) :
            ss << "BREAKPOINT type specified, but no breakpoints given\n" << "\nPROCESS: " << errorNode->Parent()->ToElement()->Attribute("name") << "\nPARAM: " << errorNode->Attribute("name");
            break;
        case(ChainError::Missing_BREAKPOINT_TIME) :
            ss << "BREAKPOINT missing 'time'\n" << "\nPROCESS: " << errorNode->Parent()->ToElement()->Attribute("name") << "\nPARAM: " << errorNode->Attribute("name");
            break;
        case(ChainError::Missing_BREAKPOINT_VALUE) :
            ss << "BREAKPOINT mising 'value'\n" << "\nPROCESS: " << errorNode->Parent()->ToElement()->Attribute("name") << "\nPARAM: " << errorNode->Attribute("name");
            break;
        default: // should hopefully never reach here
            ss << "Unknown Chain error";
        }

        return ChainException(ss.str().c_str());
    }

    static EngineException create(EngineError errorCode)
    {
        std::stringstream ss;
        ss << "ERROR IN ENGINE: ";
        switch (errorCode)
        {
        case(EngineError::FailedToOpen_XML) :
            ss << "Could not open XML file";
            break;
        case(EngineError::Element_ROOT_MISSING) :
            ss << "Root element missing from XML - is this a broken/empty file?";
            break;
        case(EngineError::Element_LINK_MISSING) :
            ss << "LINK is not the root XML element - is this a linkEngine file?";
            break;
        default: // should hopefully never reach here
            ss << "unknown error";
        }

        return EngineException(ss.str().c_str());
    }
};

#endif // LINK_EXCEPTIONFACTORY_H
