#ifndef LINK_UTILS_H
#define LINK_UTILS_H

#include <cmath>
#include <vector>
#include <string>

namespace linkEngine
{
    const double PI = 4 * atan(1.f);

    inline void tokenise(const std::string& toTokenise, std::vector<std::string>* tokens, const char* delim)
    {
        int start = 0, end = 0;

        while ((end = toTokenise.find(delim, start)) != std::string::npos)
        {
            std::string token = toTokenise.substr(start, end - start);
            if (token != "")
                tokens->push_back(token);
            start = end + 1;
        }

        std::string token = toTokenise.substr(start);
        if (token != "")
            tokens->push_back(token);
    }

    class Phasor
    {
        double m_phase;
        double m_incr;
        float m_mul, m_add;
        int m_sampleRate;

    public:
        Phasor(const float freq, const int sampleRate = 44100, const float mul = 1, const float add = 0)
            : m_phase(0.0), m_sampleRate(sampleRate), m_mul(mul), m_add(add)
        {
            setFreq(freq);
        }

        void setFreq(const float freq)
        {
            m_incr = freq / static_cast<double>(m_sampleRate - 1);
        }

        void tick()
        {
            m_phase += m_incr;
            if (m_phase > 1)
            {
                m_phase -= 1;
            }
            else if (m_phase < 0)
            {
                m_phase += 1;
            }
        }

        double get() const
        {
            return m_phase * m_mul + m_add;
        }
    };

    template<class T> class Table
    {
        std::vector<T> m_table;
    public:
        Table(int size)
            : m_table(size, 0)
        {
            fillSine();
        }

        void fillSine(float freq = 1, float mul = 1, float add = 0)
        {
            for (int i = 0; i < m_table.size(); ++i)
            {
                m_table[i] = mul * sin(2 * PI * freq * i / (m_table.size() - 1)) + add;
            }
            m_table[m_table.size() - 1] = m_table[0];
        }

        T get(int index) const
        {
            T value = 0;

            if (index < 0)
                value = m_table.front();
            else if (index > m_table.size())
                value = m_table.back();
            else
                value = m_table[index];

            return value;
        }

        T get_linear(float index) const
        {
            int lowerBound = static_cast<int>(index);
            int upperBound = lowerBound + 1;

            T value = 0;

            if (lowerBound < 0)
                value = m_table.front();
            else if (upperBound > m_table.size())
                value = m_table.back();
            else
            {
                T left = m_table[lowerBound];
                T right = m_table[upperBound];

                T frac = (index - lowerBound) / 1.f;
                value = left + ((right - left) * frac);
            }
            return value;
        }

        int size() const
        {
            return m_table.size() - 1;
        }
    };

} // linkEngine namespace

#endif // LINK_UTILS_H