SOC_Param
{
	var <name, <argName, <>min, <>max, curValue, parent, restoreVal;

	*new { | name, argName, min, max, initValue |
		^super.newCopyArgs(name, argName, min, max, initValue);
	}

	setParent { | parentProcess|
		parent = parentProcess;
	}

	value {
		^curValue.clip(min, max);
	}

	value_ { | v |
		curValue = v.clip(min,max);
		if(parent.synth != nil)
		{
			if(parent.synth.isPlaying)
			{
				parent.synth.set(argName, curValue);
			}
		}
	}

	vary {
		if(restoreVal == nil) { restoreVal = this.value; };
		this.value_(rrand(min.asFloat, max.asFloat));
	}

	restore {
		if(restoreVal != nil)
		{
			this.value_(restoreVal);
			restoreVal = nil;
		}
	}

	printOn { | stream |
		stream << "SOC_Param(" << name << "," << this.value << ")";
	}

	view {
		var spec = ControlSpec(this.min, this.max);
		var slider, numBox;

		^[
			StaticText().string_(this.name),
			slider = Slider().orientation_(\horizontal).action_({ | caller | numBox.valueAction_(spec.map(caller.value)); }).value_(spec.unmap(this.value)),
			numBox = NumberBox().action_({ | caller | slider.value_(spec.unmap(caller.value)); this.value_(caller.value); }).value_(this.value);
		]
	}
}

SOC_SilentParam : SOC_Param
{
	*new { | name, min, max, initValue |
		^super.new(name, nil, min, max, initValue)
	}

	value_{ | v |
		curValue = v.clip(min, max);
	}
}

// Abstract Process
SOC_Process
{
	var <name, compileName, <params, <synth;

	*new { | name, compileName |
		^super.new.init(name, compileName);
	}

	init { | n, cn |
		name = n;
		compileName = cn;
		params = List();
	}

	clear {
		if(synth != nil)
		{
			if(synth.isPlaying)
			{
				synth.free;
			}
		};
		synth = nil;
	}

	vary { | paramName |
		if(paramName != nil)
		{
			var p = this.getParam(paramName);
			if(p != nil)
			{
				p.vary;
			}
		}
	}

	restore { | paramName |
		if(paramName != nil)
		{
			var p = this.getParam(paramName);
			if(p != nil)
			{
				p.restore;
			}
		}
	}

	addParam { | name, argName, min, max, init |
		params.add(SOC_Param(name, argName, min, max, init).setParent(this));
	}

	getParam { | name |
		^params.select({|param| param.name == name})[0];
	}

	addSilentParam { | name, min, max, init |
		params.add(SOC_SilentParam(name, min, max, init));
	}

	process { | parent |
		this.subclassResponsibility(thisMethod);
	}

	storeOn { | stream |
		var cs;

		cs = compileName ++ "(";

		params.do({|param|
			cs = cs + param.value + ",";
		});

		cs = cs.drop(-1);
		cs = cs ++ ")";

		stream << cs;
	}

	view {
		var view;
		var layout;

		view = View();
		view.layout_(layout = GridLayout());

		params.do({ | param, row |
			param.view.do({ | item, col |
				layout.add(item, row, col);
			})
		});

		^view;
	}
}

SOC_Player : SOC_Process
{
	*new { | startPos=0, rate=1, gain=1, wobbleFreq=1, wobbleWidth=0 |
		^super.new.init(startPos, rate, gain, wobbleFreq, wobbleWidth);
	}

	init { | startPos, rate, gain, wobbleFreq, wobbleWidth |
		super.init("Player", "SOC_Player");

		this.addParam("Start Pos", \startPos, 0, 1, startPos);
		this.addParam("Rate", \rate, 0.01, 20, rate);
		this.addParam("Gain", \gain, 0, 2, gain);
		this.addParam("Wobble Freq", \wobbleFreq, 0, 100, wobbleFreq);
		this.addParam("Wobble Width", \wobbleWidth, 0, 2, wobbleWidth);
	}

	process { | parent |
		synth = { | startPos, rate, gain, wobbleFreq, wobbleWidth |
			PlayBuf.ar(parent.buffer.numChannels, parent.buffer, rate + SinOsc.ar(wobbleFreq, 0, wobbleWidth), startPos: startPos * BufFrames.kr(parent.buffer), doneAction: 2) * gain;
		}.play(
			SOC_SoundObject.processGroup,
			parent.processBus,
			addAction: \addToTail,
			args: [
				\startPos, this.getParam("Start Pos").value,
				\rate, this.getParam("Rate").value,
				\gain, this.getParam("Gain").value,
				\wobbleWidth, this.getParam("Wobble Width").value,
				\wobbleFreq, this.getParam("Wobble Freq").value
			]
		);
		NodeWatcher.register(synth, true);
	}
}

SOC_RingMod : SOC_Process
{
	*new { | freq = 10, width = 1|
		^super.new.init(freq, width);
	}

	init { | freq, width |
		super.init("Ring Modulation", "SOC_RingMod");

		this.addParam("Freq", \freq, 0, 1000, freq);
		this.addParam("Width", \width, 0, 1, width);
	}

	process { | parent |
		synth = { | freq, width |
			ReplaceOut.ar(parent.processBus, parent.processBus.ar() * SinOsc.ar(freq, 0, width * 0.5, 0.5));
		}.play(SOC_SoundObject.processGroup, parent.processBus, addAction: \addToTail, args: [\freq, this.getParam("Freq").value, \width, this.getParam("Width").value]);
		NodeWatcher.register(synth, true);
	}
}

SOC_Gain : SOC_Process
{
	*new { | gain = 0.5 |
		^super.new.init(gain);
	}

	init { | gain |
		super.init("Gain", "SOC_Gain");

		this.addParam("Gain", \gain, 0, 2, gain);
	}

	process { | parent |
		synth = { | gain |
			ReplaceOut.ar(parent.processBus, parent.processBus.ar() * gain);
		}.play(SOC_SoundObject.processGroup, parent.processBus, addAction: \addToTail, args: [\gain, this.getParam("Gain").value]);
		NodeWatcher.register(synth, true);
	}
}

SOC_Pan : SOC_Process
{
	*new { | pos = 0, randSpeed=10, randWidth=0 |
		^super.new.init(pos, randSpeed, randWidth);
	}

	init { | pos, randSpeed, randWidth |
		super.init("Pan", "SOC_Pan");

		this.addParam("Position", \pos, -1, 1, pos);
		this.addParam("Random Speed", \randSpeed, 0.01, 20, randSpeed);
		this.addParam("Random Width", \randWidth, 0, 1, randWidth);
	}

	process { | parent |
		synth = { | pos, randSpeed, randWidth |
			var sig, pan;

			sig = parent.processBus.ar();

			pan = pos + ((1 - pos.abs) * LFNoise1.kr(randSpeed, randWidth));

			ReplaceOut.ar(parent.processBus, Pan2.ar(sig[0],pan));

		}.play(SOC_SoundObject.processGroup, parent.processBus, addAction: \addToTail, args: [\pos, this.getParam("Position").value, \randSpeed, this.getParam("Random Speed").value, \randWidth, this.getParam("Random Width").value]);
		NodeWatcher.register(synth, true);
	}
}

SOC_Balance : SOC_Process
{
	*new { | pos = 0, randSpeed=10, randWidth=0 |
		^super.new.init(pos, randSpeed, randWidth);
	}

	init { | pos, randSpeed, randWidth |
		super.init("Balance", "SOC_Balance");

		this.addParam("Position", \pos, -1, 1, pos);
		this.addParam("Random Speed", \randSpeed, 0.01, 20, randSpeed);
		this.addParam("Random Width", \randWidth, 0, 1, randWidth);
	}

	process { | parent |
		synth = { | pos, randSpeed, randWidth |
			var sig, pan;

			sig = parent.processBus.ar();

			pan = pos + ((1 - pos.abs) * LFNoise1.kr(randSpeed, randWidth));

			ReplaceOut.ar(parent.processBus, Pan2.ar(sig[0],pan));

		}.play(SOC_SoundObject.processGroup, parent.processBus, addAction: \addToTail, args: [\pos, this.getParam("Position").value, \randSpeed, this.getParam("Random Speed").value, \randWidth, this.getParam("Random Width").value]);
		NodeWatcher.register(synth, true);
	}
}


SOC_Stack : SOC_Process
{
	var stackGroup;

	*new { | numInStack=2, freqCoef=1.5, freqOffset=0, timeOffset=0, gain=0.5 |
		^super.new.init(numInStack, freqCoef, freqOffset, timeOffset, gain);
	}

	init { | numInStack, freqCoef, freqOffset, timeOffset, gain |
		super.init("Stack", "SOC_Stack");

		this.addSilentParam("Stack Size", 1, 800, numInStack);
		this.addSilentParam("Freq Coef", 0.1, 10, freqCoef);
		this.addParam("Freq Offset", \freqOffset, 0, 100, freqOffset);
		this.addSilentParam("Time Offset", 0, 100, timeOffset);
		this.addParam("Gain", \gain, 0, 10, gain);
	}

	process { | parent |
		if(stackGroup != nil) { if(stackGroup.isPlaying) { stackGroup.freeAll; } };
		stackGroup = Group(SOC_SoundObject.processGroup);
		synth = stackGroup;
		synth.isPlaying = true;

		Routine({
			this.getParam("Stack Size").value.asInteger.geom(1, this.getParam("Freq Coef").value).do({ | rate |
				{ | freqOffset, gain |
					PlayBuf.ar(parent.buffer.numChannels, parent.buffer, freqOffset + rate, doneAction: 2) * gain;
				}.play(stackGroup, parent.processBus, addAction: \addToTail, args: [\freqOffset, this.getParam("Freq Offset").value, \gain, this.getParam("Gain").value]);
				this.getParam("Time Offset").value.wait;
			});
		}).play;
	}
}

SOC_Granular : SOC_Process
{
	*new { | grainRate=10, grainSize=0.01, grainPos=0, grainPosRand=0, playbackRate=1, playbackRateRand=0 |
		^super.new.init(grainRate, grainSize, grainPos, grainPosRand, playbackRate, playbackRateRand);
	}

	init { | grainRate, grainSize, grainPos, grainPosRand, playbackRate, playbackRateRand |
		super.init("Granular", "SOC_Granular");

		this.addParam("Grain Rate", \grainRate, 0.01, 20, grainRate);
		this.addParam("Grain Size", \grainSize, 0.001, 2, grainSize);
		this.addParam("Grain Pos", \grainPos, 0, 1, grainPos);
		this.addParam("Grain Pos Rand", \grainPosRand, 0, 1, grainPosRand);
		this.addParam("Playback Rate", \pbRate, 0.01, 20, playbackRate);
		this.addParam("Playback Rate Rand", \pbRateRand, 0, 1, playbackRateRand);
	}

	process { | parent |
		synth = { | grainRate, grainSize, grainPos, grainPosRand, pbRate, pbRateRand |
			var sig;

			// parent.buffers is array of mono buffers - multichannel expansion should do what we want
			sig = GrainBuf.ar(1, Impulse.ar(grainRate), grainSize, parent.buffers, pbRate + (LFNoise1.kr(20).range(0,10) * pbRateRand), grainPos + (LFNoise1.kr(20).range(0,1) * grainPosRand), 2);

			Out.ar(parent.processBus, sig);
		}.play(
			SOC_SoundObject.processGroup,
			parent.processBus,
			addAction: \addToTail,
			args: [
				\grainRate, this.getParam("Grain Rate").value,
				\grainSize, this.getParam("Grain Size").value,
				\grainPos, this.getParam("Grain Pos").value,
				\grainPosRand, this.getParam("Grain Pos Rand").value,
				\pbRate, this.getParam("Playback Rate").value,
				\pbRateRand, this.getParam("Playback Rate Rand").value,
			]
		);
		NodeWatcher.register(synth, true);
	}
}

SOC_BPF : SOC_Process
{
	*new { | freq=440, rq=1, amp=1 |
		^super.new.init(freq, rq, amp);
	}

	init { | freq, rq, amp |
		super.init("BPF", "SOC_BPF");

		this.addParam("Freq", \freq, 10, 15000, freq);
		this.addParam("rQ", \rq, 0.001, 2, rq);
		this.addParam("Amp", \amp, 0, 5, amp);
	}

	process { | parent |
		synth = { | freq, rq, amp |
			ReplaceOut.ar(parent.processBus, BPF.ar(parent.processBus.ar(), freq, rq, amp));
		}.play(SOC_SoundObject.processGroup, parent.processBus, addAction: \addToTail, args: [\freq, this.getParam("Freq").value, \rq, this.getParam("rQ").value, \amp, this.getParam("Amp").value]);
		NodeWatcher.register(synth, true);
	}
}

SOC_Comb : SOC_Process
{
	*new { | note=60, decayTime=1,amp=1 |
		^super.new.init(note, decayTime, amp);
	}

	init { | note, decayTime, amp |
		super.init("Comb", "SOC_Comb");

		this.addParam("Note", \note, 0, 127, note);
		this.addParam("Decay", \decayTime, 0, 2, decayTime);
		this.addParam("Amp", \amp, 0, 1, amp);
	}

	process { | parent |
		synth = { | note, decayTime, amp|
			var sig;

			sig = CombC.ar(parent.processBus.ar(), 0.2, 1 / note.midicps, decayTime);

			ReplaceOut.ar(parent.processBus, sig * amp);
		}.play(SOC_SoundObject.processGroup, parent.processBus, addAction: \addToTail, args: [\note, this.getParam("Note").value, \decayTime, this.getParam("Decay").value, \amp, this.getParam("Amp").value]);
		NodeWatcher.register(synth, true);
	}
}

SOC_Envelope : SOC_Process
{
	*new { | attack=0.1, sustainTime=1, sustainLevel=0.7 release=0.1, gate=0 |
		^super.new.init(attack, sustainTime, sustainLevel, release, gate);
	}

	init { | attack, sustainTime, sustainLevel, release, gate |
		super.init("Envelope", "SOC_Envelope");

		this.addParam("Attack Time", \attack, 0, 20, attack);
		this.addParam("Sustain Time", \sustainTime, 0, 20, sustainTime);
		this.addParam("Sustain Level", \sustainLevel, 0, 1, sustainLevel);
		this.addParam("Release Time", \releaseTime, 0, 20, release);
		this.addParam("Gate", \gate, -1, 1, gate);
	}

	trig {
		var gate = this.getParam("Gate");

		gate.value = 0;
		gate.value = 1;
	}

	process { | parent |
		synth = { | attack, sustainTime, sustainLevel, release, gate |
			var sig, env;
			sig = parent.processBus.ar();

			env = EnvGen.kr(
				Env.linen(attack, sustainTime, release, sustainLevel),
				gate
			);

			ReplaceOut.ar(parent.processBus, sig * env);
		}.play(
			SOC_SoundObject.processGroup,
			parent.processBus,
			addAction: \addToTail,
			args: [
				\attack, this.getParam("Attack Time").value,
				\sustainTime, this.getParam("Sustain Time").value,
				\sustainLevel, this.getParam("Sustain Level").value,
				\release, this.getParam("Release Time").value,
				\gate, this.getParam("Gate").value
			]
		);

		NodeWatcher.register(synth, true);
	}
}